/*
 * mshplot.c: a simple plotter of mesh data
 *
 * usage: mshplot [-r] mesh_data 
 *
 * -r: reverse surface
 *
 * mesh_data format:
 *   ugrid * vgrid
 *   xval[0][0]  yval[0][0]  zval[0][0]
 *   xval[1][0]  yval[1][0]  zval[1][0]
 *   ...
 *
 * 1999/03/29, tsakurai@rimath.saitama-u.ac.jp
 * 1999/08/14, add Postscript-Out
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/glut.h>
#include "vector.h" /* typedef struct {double x,y,z;} Vector; */
#include "bmp.c"    /* SaveImageAs()                          */
#include "ps.c"     /* DrawPS()                               */

int width = 500, height = 500;

typedef struct _Mesh {
    Vector **vx;
    Vector **vn;
    int ugrid, vgrid;
} Mesh;

/* main object of mshplot */
Mesh *mesh;
char data_name[100] = "mshplot";

#define DEG (3.14159265358979323846/180)

/* eye = (r, theta, phi) in spherical coord */
double r = 9, theta = 70*DEG, phi = -80*DEG;
double fovx = 20;
double eye_cent[] = {.001, .0, .0};
double dtheta = 2*DEG;

/* rotation angle */
GLfloat angle  = 0.0;
GLfloat dangle = 1.0; /* in degree */

/* scale */
GLfloat scale = 0., zmag = 1.;

/* flags */
int smooth_shading = 1;
int animation = 0;
int wire  = 0;
int frame = 3; /* 2-bits mask */
int stride= 1; /* grid stride */
int axis  = 7; /* 3-bits mask */
int mag   = 0; /* magnification */
int qclip = 0; /* quarter clipping */
int psout = 0; /* out to Postscript-file */
int light = 3; /* 2-bits mask */
int clipping = 0;
int reverse  = 0;
int oblique  = 0;
int bgwhite  = 0;
int xrot = 0;
int zrot = 0;

/* material */
GLfloat material_color[]  = {.7, .6, .8, 1.};
GLfloat specular_color[]  = {.4, .4, .4, 1.};
GLfloat shininess = .6*128.0;
GLfloat mat_back_color[]  = {.45, .17, .3, 1.};
GLfloat frame_color[]     = {.0, .0, .0, 1.};

/* lights */
GLfloat light0_ambient [] = {.1, .1, .1, 1.};
GLfloat light0_diffuse [] = {.8, .8, .8, 1.};
GLfloat light0_specular[] = {.5, .5, .5, 1.};
GLfloat light0_position[] = {-15,-5, 20, 0.};
GLfloat light1_diffuse [] = {.4, .4, .4, 1.};
GLfloat light1_specular[] = {.4, .4, .4, 1.};
GLfloat light1_position[] = { 5, -5, -2, 0.};

/***************************************************
 * setupMash
 --------------------------------------------------*/
void calcNormal(void)
{
    int cylinder = 0;
    Vector **vx = mesh->vx, **vn = mesh->vn;
    int NI = mesh->ugrid, NJ = mesh->vgrid;
    int i, j;
    
    if (VDist(&vx[0][5], &vx[NI-1][5]) < 1e-8)  cylinder |= 1;
    if (VDist(&vx[5][0], &vx[5][NJ-1]) < 1e-8)  cylinder |= 2;

    for (j = 0; j < NJ; j++)
	for (i = 0; i < NI; i++) {
	    static Vector O = {0, 0, 0}; vn[i][j] = O;
	}
    for (j = 0; j < NJ-1; j++)
	for (i = 0; i < NI-1; i++) {
	    static Vector v0, v1, v2, n0, n1;
	    /*
	     * vx[i][j+1]  vx[i+1][j+1]
	     *  v0 |  v1 /  v2 |
	     * vx[i][j]    vx[i+1][j]
	     */
	    VSub(&v0, &vx[i][j],     &vx[i][j+1]);
	    VSub(&v1, &vx[i+1][j+1], &vx[i][j]);
	    VSub(&v2, &vx[i+1][j],   &vx[i+1][j+1]);

	    /* Face0: vx[i][j+1]-vx[i][j]-vx[i+1][j+1] */
	    VCross(&n0, &v0, &v1);
	    VAdd(&vn[i][j+1],  &vn[i][j+1],  &n0);
	    VAdd(&vn[i][j],    &vn[i][j],    &n0);
	    VAdd(&vn[i+1][j+1],&vn[i+1][j+1],&n0);

	    /* Face1: vx[i+1][j+1]-vx[i][j]-vx[i+1][j] */
	    VCross(&n1, &v2, &v1);
	    VAdd(&vn[i+1][j+1],&vn[i+1][j+1],&n1);
	    VAdd(&vn[i][j],    &vn[i][j],    &n1);
	    VAdd(&vn[i+1][j],  &vn[i+1][j],  &n1);
	}
    if (cylinder & 1) 
	for (j = 0; j < NJ; j++) {
	    VAdd(&vn[0][j], &vn[0][j], &vn[NI-1][j]);
	    vn[NI-1][j] = vn[0][j];
	}
    if (cylinder & 2) 
	for (i = 0; i < NI; i++) {
	    VAdd(&vn[i][0], &vn[i][0], &vn[i][NJ-1]);
	    vn[i][NJ-1] = vn[i][0];
	}
    for (j = 0; j < NJ; j++) 
	for (i = 0; i < NI; i++) {
	    VNormalize(&vn[i][j]);
	}
}

void setupMesh(void)
{
    double Max = 0, min = 0;
    Vector **vx = mesh->vx;
    int NI = mesh->ugrid, NJ = mesh->vgrid;
    int i, j;

    if (reverse)
	for (i = 0; i < NI/2; i++) {
	    Vector *swap = vx[i]; vx[i] = vx[NI-1-i]; vx[NI-1-i] = swap;
	}
    calcNormal();
    for (j = 0; j < NJ; j++)
	for (i = 0; i < NI; i++) {
	    Vector *v = &vx[i][j];
	    Max = (v->x > Max) ? v->x : (v->y > Max) ? v->y : Max;
	    min = (v->x < min) ? v->x : (v->y < min) ? v->y : min;
	}
    if (scale == 0.0) {
	scale = 1.0/((-min > Max) ? -min : Max);
	for (i = 0; i < mag; i++) scale/=.9;
    }
}

/***************************************************
 * rendering
 --------------------------------------------------*/
void drawSurface(void)
{
    Vector **vx = mesh->vx, **vn = mesh->vn;
    int NI = mesh->ugrid, NJ = mesh->vgrid;
    int i, j;
    
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material_color);
    glMaterialfv(GL_FRONT, GL_SPECULAR, specular_color);
    glMaterialf (GL_FRONT, GL_SHININESS, shininess);
    glMaterialfv(GL_BACK , GL_AMBIENT_AND_DIFFUSE, mat_back_color);
    for (j = 0; j < NJ-1; j++) {
	glBegin(GL_TRIANGLE_STRIP);
	for (i = 0; i <= NI-1; i++) 
	    if (!qclip || i<=(NI-1)/2&&(qclip&1) || j>=(NJ-1)/2&&(qclip&2)){
		glNormal3dv(&vn[i][j+1].x); glVertex3dv(&vx[i][j+1].x);
		glNormal3dv(&vn[i][j].x);   glVertex3dv(&vx[i][j].x);
	    }
	glEnd();
    }
}

void drawFrame(void)
{
    Vector **vx = mesh->vx, **vn = mesh->vn;
    int NI = mesh->ugrid, NJ = mesh->vgrid;
    int i, j;
    
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material_color);
    glMaterialfv(GL_FRONT, GL_SPECULAR, frame_color); /* frame_color=brack */
   
    glEnable(GL_NORMALIZE);
    if (frame & 1)
	for (j = 0; j < NJ; j += stride) {
	    glBegin(GL_LINE_STRIP);
	    for (i = 0; i < NI; i++) 
		if (!qclip||i<=(NI-1)/2&&(qclip&1)||j>=(NJ-1)/2&&(qclip&2)){
		    glNormal3dv(&vn[i][j].x);
		    glVertex3dv(&vx[i][j].x);
		}
	    glEnd();
	}
    if (frame & 2)
	for (i = 0; i < NI; i += stride) {
	    glBegin(GL_LINE_STRIP);
	    for (j = 0; j < NJ; j++) 
		if (!qclip||i<=(NI-1)/2&&(qclip&1)||j>=(NJ-1)/2&&(qclip&2)){
		    glNormal3dv(&vn[i][j].x);
		    glVertex3dv(&vx[i][j].x);
		}
	    glEnd();
	}
}

void drawAxis(void)
{
    glDisable(GL_LIGHTING);
    glBegin(GL_LINES);
    if (axis & (1<<0)) {
	glColor3f(.5,.0,.2);
	glVertex3d(-1.7,.0,.0); glVertex3d(1.7,.0,.0);
    }
    if (axis & (1<<1)) {
	glColor3f(.0,.3,.2);
	glVertex3d(.0,-1.7,.0); glVertex3d(.0,1.7,.0);
    }
    if (axis & (1<<2)) {
	glColor3f(.1,.0,.3);
	glVertex3d(.0,.0,-1.3); glVertex3d(.0,.0,1.3);
    }
    glEnd();
    glEnable(GL_LIGHTING);
}

/***************************************************
 * glutCallback
 --------------------------------------------------*/
void myinit(void)
{
    /* set light properties */
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, light0_ambient);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
	glLightfv(GL_LIGHT0, GL_DIFFUSE,  light0_diffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light0_specular);
    glEnable(GL_LIGHT1);
	glLightfv(GL_LIGHT1, GL_DIFFUSE,  light1_diffuse);
	glLightfv(GL_LIGHT1, GL_SPECULAR, light1_specular);

    /* Z-Buffer */
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glEnable(GL_POLYGON_OFFSET_FILL);
    glPolygonOffset(1.2, 1.0);

    glEnable(GL_NORMALIZE);
}

void draw(void)
{
    /* set camera */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    if (oblique) {
	double w = 1.5*r*tan(fovx*DEG/2), h = w*height/width;
	angle = animation = 0;
	glFrustum(-1.5*r*(90*DEG+phi)  - w/2, -1.5*r*(90*DEG+phi)  + w/2,
		  -1.5*r*(90*DEG-theta)- h/2, -1.5*r*(90*DEG-theta)+ h/2,
		   1.5*r, 3.*r);
    } else { 
	gluPerspective(fovx*height/width, (double)width/height, .8*r, 1.3*r);
    }
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    if (oblique) {
	gluLookAt(2*r*(90*DEG+phi) + eye_cent[0], -2*r,
		  2*r*(90*DEG-theta) + eye_cent[2],
		  2*r*(90*DEG+phi) + eye_cent[0],  0.0,
		  2*r*(90*DEG-theta) + eye_cent[2],
		  0, 0, 1);
    } else {
	gluLookAt(r*sin(theta)*cos(phi),r*sin(theta)*sin(phi),r*cos(theta),
		  eye_cent[0], eye_cent[1], eye_cent[2],
		  -cos(theta)*cos(phi), -cos(theta)*sin(phi), sin(theta));
    }

    /* put light */
    glLightfv(GL_LIGHT0, GL_POSITION, light0_position);
    glLightfv(GL_LIGHT1, GL_POSITION, light1_position);
    (light & 1) ? glEnable(GL_LIGHT0) : glDisable(GL_LIGHT0);
    (light & 2) ? glEnable(GL_LIGHT1) : glDisable(GL_LIGHT1);

    /* clipping */
    if (clipping) {
	GLdouble clip[4] = {-1/.3, 0., -1/.3, 1};
	glClipPlane((GLenum) GL_CLIP_PLANE0, clip);
	glEnable((GLenum) GL_CLIP_PLANE0);
    }
    
    /* draw objects */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glPushMatrix(); {
	glRotated(angle, 0, 0, 1);
	glRotated(xrot*30, 1, 0, 0);
	glRotated(zrot*30, 0, 0, 1);
	glPushMatrix(); {
	    glScaled(scale, scale, scale*zmag);
	    if (wire) {
		drawFrame();
	    } else {
		glShadeModel(smooth_shading ? GL_SMOOTH : GL_FLAT);
		drawSurface();
		if (frame) {
		    glDisable(GL_LIGHTING);
		    glColor4fv(frame_color);
		    drawFrame();
		    glEnable(GL_LIGHTING);
		}
	    }
	}
	glPopMatrix();
	if (axis) {
	    GLint mode; glGetIntegerv(GL_RENDER_MODE, &mode);
	    if (mode == GL_FEEDBACK) glPassThrough(1);
	    if (clipping) glDisable((GLenum) GL_CLIP_PLANE0);
	    drawAxis();
	}
    }
    glPopMatrix();
    glutSwapBuffers();
}

void idle(void)
{
    if (psout) {
	char fname[100]; sprintf(fname,"%s.ps", data_name);
	DrawPS(fname,wire,frame,axis);
	exit(0);
    }
    if (animation) {
	angle += dangle;
	angle = (angle > 360.0) ? angle - 360.0 : angle;
	glutPostRedisplay();
    }
}

void reshape(int w, int h)
{
    glViewport(0, 0, width=w, height=h);
    glutPostRedisplay();
}

void action(int k)
{
    char fname[100]; 

    switch (k) {
    case'q': exit(0);
    case'1': light ^= 1; break;
    case'2': light ^= 2; break;
    case'a': axis = 7*!axis; break;
    case'j': axis^= (1<<0); break;
    case'k': axis^= (1<<1); break;
    case'l': axis^= (1<<2); break;
    case'c': animation = !animation; break;
    case'f': frame = 3*!frame; wire = 0; break;
    case',': frame^= 1; break;
    case'.': frame^= 2; break;
    case'u': r *=.9; break;
    case'd': r /=.9; break;
    case'>': fovx *=.9; break;
    case'<': fovx /=.9; break;
    case'r': reverse = 1; setupMesh(); break;
    case'C': clipping = !clipping; break;
    case'4': qclip = 3*!qclip; break;
    case'5': qclip^= 2; break;
    case'6': qclip^= 1; break;
    case'o': oblique = !oblique; break;
    case'+': scale /=.9; break;
    case'-': scale *=.9; break;
    case';': scale *=.9; break;
    case'z': zmag /= .9; break;
    case'Z': zmag *= .9; break;
    case'g': zmag  = 1.; frame = 3;  wire = 0; break;
    case'^': xrot = (++xrot%12); break;
    case'|': zrot = (++zrot%12); break;
    case's': smooth_shading = !smooth_shading; wire = 0; break;
    case'w': (wire=!wire)*frame == 0 ? frame = 3 : 0;  break;
    case'*': dangle *= 2; dtheta *= 2; break;
    case'/': dangle /= 2; dtheta /= 2; break;
    case'b': 
	if ((bgwhite=!bgwhite)) glClearColor(1,1,1,1);
	else glClearColor(0,0,0,0);
	break;
    case'B': 
	sprintf(fname,"%s.bmp", data_name);
	draw(); glReadBuffer(GL_FRONT); SaveImageAs(fname);
	break;
    case'S':
	sprintf(fname,"%s.ps",  data_name); DrawPS(fname, wire, frame, axis);
	break;
    case'': eye_cent[2] += dtheta; break;
    case'': eye_cent[2] -= dtheta; break;
    case'': eye_cent[0] += dtheta; break;
    case'': eye_cent[0] -= dtheta; break;
    case 1001: case 1002: case 1003: case 1004: /* for grid_stride_menu */
    case 1005: case 1010: case 1020:
 	stride = k - 1000;
	break;
    default:
	return;
    }
    glutPostRedisplay();
}

void key(unsigned char k, int x, int y) { action((int) k); }

void skey(int k, int x, int y)
{
    switch (k) {
    case GLUT_KEY_UP:    theta -= dtheta; break;
    case GLUT_KEY_DOWN:  theta += dtheta; break;
    case GLUT_KEY_LEFT:  phi   -= dtheta; break;
    case GLUT_KEY_RIGHT: phi   += dtheta; break;
    default: return;
    }
    glutPostRedisplay();
}

void makeMenus(void)
{
    int draw_mode, grid_stride;
    char bmp_out[100], ps_out[100];

    sprintf(bmp_out, "Save Image(%s.bmp)", data_name);
    sprintf(ps_out,  "PS Out to (%s.ps)",  data_name);
    draw_mode = glutCreateMenu(action); {
	glutAddMenuEntry(" -Shading (Smooth/Flat)", 's');
	glutAddMenuEntry(" -Toggle Mesh",           'f');
	glutAddMenuEntry(" -Toggle Wire Frame",     'w');
	glutAddMenuEntry(" -Toggle Axis",           'a');
	glutAddMenuEntry(" -Toggle Clipping",       'C');
    }
    grid_stride = glutCreateMenu(action); {
	glutAddMenuEntry(" 1", 1001);
	glutAddMenuEntry(" 2", 1002);
	glutAddMenuEntry(" 3", 1003);
	glutAddMenuEntry(" 4", 1004);
	glutAddMenuEntry(" 5", 1005);
	glutAddMenuEntry("10", 1010);
	glutAddMenuEntry("20", 1020);
    }
    glutCreateMenu(action); {
	glutAddSubMenu("Drawing Mode",  draw_mode);
	glutAddMenuEntry("Animation(on/off)", 'c');
	glutAddMenuEntry("Light1  (on/off)",  '1');
	glutAddMenuEntry("Light2  (on/off)",  '2');
	glutAddMenuEntry("Close up",          'u');
	glutAddMenuEntry("Move away",         'd');
	glutAddMenuEntry("Reverse Surface",   'r');
	glutAddSubMenu("Grid Stride", grid_stride);
	glutAddMenuEntry(bmp_out,             'B');
	glutAddMenuEntry(ps_out,              'S');
	glutAddMenuEntry("Quit",              'q');
    }
    glutAttachMenu(GLUT_RIGHT_BUTTON);
}

/***************************************************
 * main
 --------------------------------------------------*/
int main(int argc, char *argv[])
{
    void parseArgs(int argc, char *argv[], char *inFileName);
    Mesh *readMesh(FILE *);
    FILE *fp;

    parseArgs(argc, argv, data_name);
    if (strcmp(data_name, "mshplot")) {
	if ((fp = fopen(data_name, "r")) == NULL) {
	    fprintf(stderr, "cannot open file: %s\n", data_name); exit (1);
	}
	strtok(data_name, ".");
	mesh = readMesh(fp);
    } else
	mesh = readMesh(stdin);    

    /* setup object */
    setupMesh();

    /* initialize */
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
    glutInitWindowPosition(0, 0);
    glutInitWindowSize(width, height);
    glutCreateWindow(data_name);
    myinit();

    /* regist call-back functions */
    glutDisplayFunc(draw);
    glutReshapeFunc(reshape);
    glutKeyboardFunc(key);
    glutSpecialFunc(skey);
    glutIdleFunc(idle);
    makeMenus();

    glutMainLoop();
    return 0;
}

/***************************************************
 * parseArgs
 --------------------------------------------------*/
void usage(void)
{
    fprintf(stderr, "mshplot mesh_file [-g WxH] [-r] [-w] [-o] [-ps]\\\n"
	    "        [-frame mask] [-axis mask] [-light mask]"
	    " [-stride step] [-mag n]\\\n"
	    "        [-ecf <x,y,z>,<fovx>] \\\n");
    fprintf(stderr, "  -r : reverse surface\n");
    fprintf(stderr, "  -w : wire frame\n");
    fprintf(stderr, "  -o : oblique view\n");
    fprintf(stderr, "  -ps: draw to a Postscript file\n");
    fprintf(stderr, "  -opt <wire,frame,stride,mag,axis,qclip,oblique>\n");
    fprintf(stderr, "\n");
    exit(1);
}

#define arg1(Op,Var) \
if (strncmp(argv[idx], Op, 2) == 0) {\
    if (++idx >= argc) {\
	fprintf(stderr, "\nError in %s\n\n", argv[--idx]);\
	usage(); exit(1); }\
    Var = atoi(argv[idx]);\
    continue;\
}

void parseArgs(int argc, char *argv[], char *inFileName)
{
    int idx;
    for (idx = 1; idx < argc; idx++) {
	if (*argv[idx] != '-') { /* must be a file name */
	    strcpy(inFileName, argv[idx]); continue;
	}
	if (strcmp(argv[idx], "-g") == 0) {
	    if (sscanf(argv[++idx], "%dx%d", &width, &height) != 2) 
		fprintf(stderr,"format: -g <width>x<height>\n"), exit(1);
	    continue;
	}
	if (strncmp(argv[idx], "-ecf", 4) == 0) {
	    if (sscanf(argv[++idx], "%lf,%lf,%lf,%lf",
		       &eye_cent[0], &eye_cent[1], &eye_cent[2], &fovx) != 4) 
		fprintf(stderr,"format: -ecf <x,y,z>,<fovx>\n"), exit(1);
	    continue;
	}
	if (strncmp(argv[idx], "-opt", 4) == 0) {
	    if (sscanf(argv[++idx], "%d,%d,%d,%d,%d,%d,%d",
		       &wire,&frame,&stride,&axis,&mag,&qclip,&oblique) != 7) 
		fprintf(stderr,
		"format: -opt  <wire,frame,stride,mag,axis,qclip,oblique>\n"),
		exit(1);
	    continue;
	}
	arg1("-frame",   frame);
	arg1("-stride",  stride);
	arg1("-mag",     mag);
	arg1("-axis",    axis);
	arg1("-light",   light);
	if (strncmp(argv[idx], "-reverse", 2) == 0) { reverse = 1; continue; }
	if (strncmp(argv[idx], "-wire",    2) == 0) { wire    = 1; continue; }
	if (strncmp(argv[idx], "-oblique", 2) == 0) { oblique = 1; continue; }
	if (strncmp(argv[idx], "-psout",   3) == 0) { psout   = 1; continue; }

	/* else: show usage */
	fprintf(stderr, "\nUnknown option: %s\n\n", argv[idx]);
	usage(); exit(1);
    }
}

/***************************************************
 * read Mesh
 --------------------------------------------------*/
static void mygets(char *linebuf, FILE *fp)
{
    for (; ;) {
	if (fgets(linebuf, sizeof(char)*256, fp) == NULL) {
	    fprintf(stderr, "cannot read any more.\n"); exit(1);
	}
	if (linebuf[0] != '#' && linebuf[0] != '\n') return;
    }
}

Mesh *readMesh(FILE *fp)
{
    static char linebuf[256];
    Mesh *mesh;
    int i, j;

    mesh = (Mesh *) malloc(sizeof(Mesh));
    do {
	mygets(linebuf, fp);
    } while (sscanf(linebuf, "%d * %d", &mesh->ugrid, &mesh->vgrid) != 2);
    mesh->vx = (Vector **) malloc(sizeof(Vector *)*mesh->ugrid);
    mesh->vn = (Vector **) malloc(sizeof(Vector *)*mesh->ugrid);
    for (i = 0; i < mesh->ugrid; i++) {
	mesh->vx[i] = (Vector *) malloc(sizeof(Vector)*mesh->vgrid);
	mesh->vn[i] = (Vector *) malloc(sizeof(Vector)*mesh->vgrid);
    }
    for (j = 0; j < mesh->vgrid; j++)
	for (i = 0; i < mesh->ugrid; i++) {
	    Vector *vx = &mesh->vx[i][j];
	    mygets(linebuf, fp);
	    if (sscanf(linebuf, "%lf %lf %lf", &vx->x, &vx->y, &vx->z) != 3)
		fputs("not in 'x y z' format.", stderr),  exit(1);
	}
    fclose(fp);
    return mesh;
}
